from typing import TYPE_CHECKING, Literal, List
from pydantic import Field, HttpUrl, validator
from ninja import Schema
from hydrothings.schemas import BaseListResponse, BaseGetResponse, BasePostBody, BasePatchBody, NestedEntity
from hydrothings.validators import allow_partial
from hydrothings.components.sensors.utils import metadata_validator

if TYPE_CHECKING:
    from hydrothings.components.datastreams.schemas import Datastream


sensorEncodingTypes = Literal[
    'application/pdf',
    'http://www.opengis.net/doc/IS/SensorML/2.0',
    'text/html'
]


class SensorFields(Schema):
    name: str
    description: str
    encoding_type: sensorEncodingTypes = Field(..., alias='encodingType')
    sensor_metadata: str = Field(..., alias='metadata')
    properties: dict = {}

    _metadata_validator = validator('sensor_metadata', allow_reuse=True, check_fields=False)(metadata_validator)


class SensorRelations(Schema):
    datastreams: List['Datastream'] = []


class Sensor(SensorFields, SensorRelations):
    pass


class SensorPostBody(BasePostBody, SensorFields):
    datastreams: List[NestedEntity] = Field(
        [], alias='Datastreams', nested_class='DatastreamPostBody'
    )


@allow_partial
class SensorPatchBody(BasePatchBody, SensorFields):
    pass


class SensorGetResponse(SensorFields, BaseGetResponse):
    datastreams_link: HttpUrl = Field(..., alias='Datastreams@iot.navigationLink')


class SensorListResponse(BaseListResponse):
    value: List[SensorGetResponse]



from hydrothings.main import SensorThingsAPI, SensorThingsEndpoint
from hydrothings.engine import SensorThingsAbstractEngine, SensorThingsRequest

__all__ = [
    "SensorThingsAPI",
    "SensorThingsAbstractEngine",
    "SensorThingsRequest",
    "SensorThingsEndpoint"
]



from typing import TYPE_CHECKING, List
from pydantic import Field, HttpUrl
from ninja import Schema
from hydrothings.schemas import BaseListResponse, BaseGetResponse, BasePostBody, BasePatchBody, NestedEntity
from hydrothings.validators import allow_partial

if TYPE_CHECKING:
    from hydrothings.components.datastreams.schemas import Datastream


class ObservedPropertyFields(Schema):
    name: str
    definition: HttpUrl
    description: str
    properties: dict = {}


class ObservedPropertyRelations(Schema):
    datastreams: List['Datastream'] = []


class ObservedProperty(ObservedPropertyFields, ObservedPropertyRelations):
    pass


class ObservedPropertyPostBody(BasePostBody, ObservedPropertyFields):
    datastreams: List[NestedEntity] = Field(
        [], alias='Datastreams', nested_class='DatastreamPostBody'
    )


@allow_partial
class ObservedPropertyPatchBody(BasePatchBody, ObservedPropertyFields):
    pass


class ObservedPropertyGetResponse(ObservedPropertyFields, BaseGetResponse):
    datastreams_link: HttpUrl = Field(..., alias='Datastreams@iot.navigationLink')


class ObservedPropertyListResponse(BaseListResponse):
    value: List[ObservedPropertyGetResponse]



from typing import TYPE_CHECKING, List, Union
from datetime import datetime
from pydantic import Field, HttpUrl
from ninja import Schema
from hydrothings.schemas import BaseListResponse, BaseGetResponse, BasePostBody, BasePatchBody, EntityId, NestedEntity
from hydrothings.validators import allow_partial

if TYPE_CHECKING:
    from hydrothings.components.things.schemas import Thing
    from hydrothings.components.locations.schemas import Location


class HistoricalLocationFields(Schema):
    time: datetime


class HistoricalLocationRelations(Schema):
    thing: 'Thing'
    locations: List['Location']


class HistoricalLocation(HistoricalLocationFields, HistoricalLocationRelations):
    pass


class HistoricalLocationPostBody(BasePostBody, HistoricalLocationFields):
    thing: Union[EntityId, NestedEntity] = Field(
        ..., alias='Thing', nested_class='ThingPostBody'
    )
    locations: List[Union[EntityId, NestedEntity]] = Field(
        ..., alias='Locations', nested_class='LocationPostBody'
    )


@allow_partial
class HistoricalLocationPatchBody(HistoricalLocationFields, BasePatchBody):
    thing: EntityId = Field(..., alias='Thing')
    locations: List[EntityId] = Field(..., alias='Locations')


class HistoricalLocationGetResponse(BaseGetResponse, HistoricalLocationFields):
    thing_link: HttpUrl = Field(..., alias='Thing@iot.navigationLink')
    historical_locations_link: HttpUrl = Field(..., alias='Locations@iot.navigationLink')


class HistoricalLocationListResponse(BaseListResponse):
    value: List[HistoricalLocationGetResponse]



from ninja import Router, Query
from django.http import HttpResponse
from hydrothings.engine import SensorThingsRequest
from hydrothings.schemas import Filters
from hydrothings.utils import entity_or_404, list_response_codes, get_response_codes
from .schemas import DatastreamPostBody, DatastreamPatchBody, DatastreamListResponse, DatastreamGetResponse


router = Router(tags=['Datastreams'])


@router.get(
    '/Datastreams',
    response=list_response_codes(DatastreamListResponse),
    by_alias=True,
    url_name='list_datastream'
)
def get_datastreams(request: SensorThingsRequest, filters: Filters = Query(...)):
    """
    Get a collection of Datastream entities.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/datastream/properties" target="_blank">\
      Datastream Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/datastream/relations" target="_blank">\
      Datastream Relations</a>
    """

    response = request.engine.list(**filters.dict())

    return 200, response


@router.get(
    '/Datastreams({datastream_id})',
    response=get_response_codes(DatastreamGetResponse),
    by_alias=True
)
def get_datastream(request: SensorThingsRequest, datastream_id: str):
    """
    Get a Datastream entity.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/datastream/properties" target="_blank">\
      Datastream Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/datastream/relations" target="_blank">\
      Datastream Relations</a>
    """

    response = request.engine.get(entity_id=datastream_id)

    return entity_or_404(response, datastream_id)


@router.post(
    '/Datastreams',
    response={201: None}
)
def create_datastream(request: SensorThingsRequest, response: HttpResponse, datastream: DatastreamPostBody):
    """
    Create a new Datastream entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/datastream/properties" target="_blank">\
      Datastream Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/datastream/relations" target="_blank">\
      Datastream Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/create-entity" target="_blank">\
      Create Entity</a>
    """

    datastream_id = request.engine.create(
        entity_body=datastream
    )

    response['location'] = request.engine.get_ref(
        entity_id=datastream_id
    )

    return 201, None


@router.patch(
    '/Datastreams({datastream_id})',
    response={204: None}
)
def update_datastream(request: SensorThingsRequest, datastream_id: str, datastream: DatastreamPatchBody):
    """
    Update an existing Datastream entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/datastream/properties" target="_blank">\
      Datastream Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/datastream/relations" target="_blank">\
      Datastream Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/update-entity" target="_blank">\
      Update Entity</a>
    """

    request.engine.update(
        entity_id=datastream_id,
        entity_body=datastream
    )

    return 204, None


@router.delete(
    '/Datastreams({datastream_id})',
    response={204: None}
)
def delete_datastream(request: SensorThingsRequest, datastream_id: str):
    """
    Delete a Datastream entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/delete-entity" target="_blank">\
      Delete Entity</a>
    """

    request.engine.delete(
        entity_id=datastream_id
    )

    return 204, None



import re
import hydrothings.schemas as core_schemas
from typing import Literal
from hydrothings import settings


def lookup_component(
        input_value: str,
        input_type: Literal['snake_singular', 'snake_plural', 'camel_singular', 'camel_plural'],
        output_type: Literal['snake_singular', 'snake_plural', 'camel_singular', 'camel_plural']
) -> str:
    """
    Accepts a component value and type and attempts to return an alternate form of the component name.

    :param input_value: The name of the component to lookup.
    :param input_type: The type of the component to lookup.
    :param output_type: The type of the component to return.
    :return output_value: The matching component name.
    """

    st_components = [
        {
            'snake_singular': re.sub(r'(?<!^)(?=[A-Z])', '_', capability['SINGULAR_NAME']).lower(),
            'snake_plural': re.sub(r'(?<!^)(?=[A-Z])', '_', capability['NAME']).lower(),
            'camel_singular': capability['SINGULAR_NAME'],
            'camel_plural': capability['NAME']
        } for capability in settings.ST_CAPABILITIES
    ]

    return next((c[output_type] for c in st_components if c[input_type] == input_value))


def list_response_codes(response_schema):
    """"""

    return {
        200: response_schema
    }


def get_response_codes(response_schema):
    """"""

    return {
        200: response_schema,
        404: core_schemas.EntityNotFound
    }


def entity_or_404(response, entity_id):
    """"""

    if response:
        return 200, response
    else:
        return 404, {'message': f'Record with ID {entity_id} does not exist.'}



from ninja import NinjaAPI, Schema
from copy import deepcopy
from django.urls import re_path
from pydantic import BaseModel
from typing import Union, Literal, Type, NewType, List
from hydrothings.backends.sensorthings.engine import SensorThingsEngine
from hydrothings.backends.odm2.engine import SensorThingsEngineODM2
from hydrothings.engine import SensorThingsAbstractEngine
from hydrothings.components.root.views import router as root_router
from hydrothings.components.datastreams.views import router as datastreams_router
from hydrothings.components.featuresofinterest.views import router as featuresofinterest_router
from hydrothings.components.historicallocations.views import router as historicallocations_router
from hydrothings.components.locations.views import router as locations_router
from hydrothings.components.observations.views import router as observations_router
from hydrothings.components.observedproperties.views import router as observedproperties_router
from hydrothings.components.sensors.views import router as sensors_router
from hydrothings.components.things.views import router as things_router


class SensorThingsAPI(NinjaAPI):

    def __init__(
            self,
            backend: Literal['sensorthings', 'odm2', None] = None,
            engine: Union[Type[NewType('SensorThingsEngine', SensorThingsAbstractEngine)], None] = None,
            endpoints: Union[List['SensorThingsEndpoint'], None] = None,
            **kwargs
    ):

        if not kwargs.get('version'):
            kwargs['version'] = '1.1'

        if kwargs.get('version') not in ['1.0', '1.1']:
            raise ValueError('Unsupported SensorThings version. Supported versions are: 1.0, 1.1')

        if backend not in ['sensorthings', 'odm2', None]:
            raise ValueError('Unsupported SensorThings backend. Supported backends are: "sensorthings", "odm2"')

        if not backend and not isinstance(engine, SensorThingsAbstractEngine):
            raise ValueError('No backend was specified, and no engine class was defined.')

        super().__init__(
            # openapi_url=f'/v{kwargs["version"]}/openapi.json',
            # docs_url=f'/v{kwargs["version"]}/docs',
            **kwargs
        )

        self.endpoints = endpoints

        if backend == 'sensorthings':
            self.engine = SensorThingsEngine
        elif backend == 'odm2':
            self.engine = SensorThingsEngineODM2
        else:
            self.engine = engine

        self.add_router('', self._build_sensorthings_router('root', root_router))
        self.add_router('', self._build_sensorthings_router('datastream', datastreams_router))
        self.add_router('', self._build_sensorthings_router('feature_of_interest', featuresofinterest_router))
        self.add_router('', self._build_sensorthings_router('historical_location', historicallocations_router))
        self.add_router('', self._build_sensorthings_router('location', locations_router))
        self.add_router('', self._build_sensorthings_router('observation', observations_router))
        self.add_router('', self._build_sensorthings_router('observed_property', observedproperties_router))
        self.add_router('', self._build_sensorthings_router('sensor', sensors_router))
        self.add_router('', self._build_sensorthings_router('thing', things_router))

    def _get_urls(self):

        urls = super()._get_urls()
        urls.append(re_path(r'^.*', lambda request: None, name='st_complex_handler'))

        return urls

    def _build_sensorthings_router(self, component, router):
        """"""

        router_copy = deepcopy(router)

        if self.endpoints:
            router_endpoints = {
                endpoint.name.split('_')[0]: endpoint
                for endpoint in self.endpoints
                if '_'.join(endpoint.name.split('_')[1:]) == component
            }

            if getattr(router_endpoints.get('get', {}), 'response_schema', None):
                for path_op in router_copy.path_operations.values():
                    for op in path_op.operations:
                        if op.view_func.__name__ == f'get_{component}':
                            op.response_models[200] = router_endpoints['get'].response_schema

        return router_copy


class SensorThingsEndpoint(BaseModel):
    name: str
    enabled: bool = True
    response_schema: Union[Type[Schema], None] = None
    body_schema: Union[Type[Schema], None] = None



from typing import TYPE_CHECKING, Literal, List
from pydantic import Field, HttpUrl
from geojson_pydantic import Feature
from ninja import Schema
from hydrothings.schemas import BaseListResponse, BaseGetResponse, BasePostBody, BasePatchBody, NestedEntity
from hydrothings.validators import allow_partial

if TYPE_CHECKING:
    from hydrothings.components.observations.schemas import Observation


featureEncodingTypes = Literal['application/geo+json']


class FeatureOfInterestFields(Schema):
    name: str
    description: str
    encoding_type: featureEncodingTypes = Field(..., alias='encodingType')
    feature: Feature
    properties: dict = {}


class FeatureOfInterestRelations(Schema):
    observations: List['Observation'] = []


class FeatureOfInterest(FeatureOfInterestFields, FeatureOfInterestRelations):
    pass


class FeatureOfInterestPostBody(BasePostBody, FeatureOfInterestFields):
    observations: List[NestedEntity] = Field(
        [], alias='Observations', nested_class='ObservationPostBody'
    )


@allow_partial
class FeatureOfInterestPatchBody(FeatureOfInterestFields, BasePatchBody):
    pass


class FeatureOfInterestGetResponse(BaseGetResponse, FeatureOfInterestFields):
    observations_link: HttpUrl = Field(..., alias='Observations@iot.navigationLink')


class FeatureOfInterestListResponse(BaseListResponse):
    value: List[FeatureOfInterestGetResponse]



from .datastreams.schemas import *
from .featuresofinterest.schemas import *
from .historicallocations.schemas import *
from .locations.schemas import *
from .observations.schemas import *
from .observedproperties.schemas import *
from .root.schemas import *
from .sensors.schemas import *
from .things.schemas import *


DatastreamRelations.update_forward_refs(
    Thing=Thing,
    Sensor=Sensor,
    ObservedProperty=ObservedProperty,
    Observation=Observation
)

FeatureOfInterestRelations.update_forward_refs(
    Observation=Observation
)

HistoricalLocationRelations.update_forward_refs(
    Thing=Thing,
    Location=Location
)

LocationRelations.update_forward_refs(
    Thing=Thing,
    HistoricalLocation=HistoricalLocation
)

ObservationRelations.update_forward_refs(
    FeatureOfInterest=FeatureOfInterest,
    Datastream=Datastream
)

ObservedPropertyRelations.update_forward_refs(
    Datastream=Datastream
)

SensorRelations.update_forward_refs(
    Datastream=Datastream
)

ThingRelations.update_forward_refs(
    Datastream=Datastream,
    Location=Location,
    HistoricalLocation=HistoricalLocation
)



from django.conf import settings

ST_VERSION = '1.1'

ST_CONFORMANCE = getattr(settings, 'ST_CONFORMANCE', [
    'http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/resource-path/resource-path-to-entities',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/request-data',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/create-entity',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/link-to-existing-entities',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/deep-insert',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/deep-insert-status-code',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/update-entity',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/delete-entity',
    'http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/historical-location-auto-creation'
])

ST_CAPABILITIES = getattr(settings, 'ST_CAPABILITIES', [
    {
        'NAME': 'Things',
        'SINGULAR_NAME': 'Thing',
        'VIEW': 'list_thing'
    },
    {
        'NAME': 'Locations',
        'SINGULAR_NAME': 'Location',
        'VIEW': 'list_location'
    },
    {
        'NAME': 'HistoricalLocations',
        'SINGULAR_NAME': 'HistoricalLocation',
        'VIEW': 'list_historical_location'
    },
    {
        'NAME': 'Datastreams',
        'SINGULAR_NAME': 'Datastream',
        'VIEW': 'list_datastream'
    },
    {
        'NAME': 'Sensors',
        'SINGULAR_NAME': 'Sensor',
        'VIEW': 'list_sensor'
    },
    {
        'NAME': 'Observations',
        'SINGULAR_NAME': 'Observation',
        'VIEW': 'list_observation'
    },
    {
        'NAME': 'ObservedProperties',
        'SINGULAR_NAME': 'ObservedProperty',
        'VIEW': 'list_observed_property'
    },
    {
        'NAME': 'FeaturesOfInterest',
        'SINGULAR_NAME': 'FeatureOfInterest',
        'VIEW': 'list_feature_of_interest'
    },
])



from abc import ABCMeta, abstractmethod
from typing import Union
from django.http import HttpRequest


class SensorThingsAbstractEngine(metaclass=ABCMeta):

    scheme: str
    host: str
    path: str

    def get_ref(self, entity_id: Union[str, None] = None, related_component: Union[str, None] = None) -> str:
        """
        Builds a reference URL for a given entity.

        :param entity_id: The id of the entity.
        :param related_component: The related component to be appended to the ref URL.
        :return: The entity's reference URL.
        """

        ref_url = f'{self.scheme}://{self.host}{self.path}'

        if entity_id is not None:
            ref_url = f'{ref_url}({entity_id})'

        if related_component is not None:
            ref_url = f'{ref_url}/{related_component}'

        return ref_url

    @abstractmethod
    def list(
            self,
            filters,
            count,
            order_by,
            skip,
            top,
            select,
            expand,
            result_format
    ) -> dict:
        """"""

        pass

    @abstractmethod
    def get(
            self,
            entity_id,
            component=None
    ) -> dict:
        """"""

        pass

    @abstractmethod
    def create(
            self,
            entity_body,
            component=None
    ) -> str:
        """"""

        pass

    @abstractmethod
    def update(
            self,
            entity_id,
            entity_body
    ) -> str:
        """"""

        pass

    @abstractmethod
    def delete(
            self,
            entity_id
    ) -> None:
        """"""

        pass


class SensorThingsRequest(HttpRequest):
    engine: SensorThingsAbstractEngine



from typing import TYPE_CHECKING, Literal, List, Union
from pydantic import Field, HttpUrl
from ninja import Schema
from hydrothings.schemas import BaseListResponse, BaseGetResponse, BasePostBody, BasePatchBody, EntityId, NestedEntity
from hydrothings.validators import allow_partial

if TYPE_CHECKING:
    from hydrothings.components.things.schemas import Thing
    from hydrothings.components.sensors.schemas import Sensor
    from hydrothings.components.observedproperties.schemas import ObservedProperty

observationTypes = Literal[
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_CategoryObservation',
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_CountObservation',
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_Measurement',
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_Observation',
    'http://www.opengis.net/def/observationType/OGC-OM/2.0/OM_TruthObservation'
]


class UnitOfMeasurement(Schema):
    name: str
    symbol: str
    definition: HttpUrl


# TODO Add validation for temporal duration types.
class DatastreamFields(Schema):
    name: str
    description: str
    unit_of_measurement: UnitOfMeasurement = Field(..., alias='unitOfMeasurement')
    observation_type: observationTypes = Field(..., alias='observationType')
    observed_area: dict = Field({}, alias='observedArea')
    phenomenon_time: Union[str, None] = Field(None, alias='phenomenonTime')
    result_time: Union[str, None] = Field(None, alias='resultTime')
    properties: dict = {}


class DatastreamRelations(Schema):
    thing: 'Thing'
    sensor: 'Sensor'
    observed_property: 'ObservedProperty'


class Datastream(DatastreamFields, DatastreamRelations):
    pass


class DatastreamPostBody(BasePostBody, DatastreamFields):
    thing: Union[EntityId, NestedEntity] = Field(
        ..., alias='Thing', nested_class='ThingPostBody'
    )
    sensor: Union[EntityId, NestedEntity] = Field(
        ..., alias='Sensor', nested_class='SensorPostBody'
    )
    observed_property: Union[EntityId, NestedEntity] = Field(
        ..., alias='ObservedProperty', nested_class='ObservedPropertyPostBody'
    )


@allow_partial
class DatastreamPatchBody(BasePatchBody, DatastreamFields):
    thing: EntityId = Field(..., alias='Thing')
    sensor: EntityId = Field(..., alias='Sensor')
    observed_property: EntityId = Field(..., alias='ObservedProperty')


class DatastreamGetResponse(BaseGetResponse, DatastreamFields):
    thing_link: HttpUrl = Field(..., alias='Thing@iot.navigationLink')
    sensor_link: HttpUrl = Field(..., alias='Sensor@iot.navigationLink')
    observed_property_link: HttpUrl = Field(..., alias='ObservedProperty@iot.navigationLink')


class DatastreamListResponse(BaseListResponse):
    value: List[DatastreamGetResponse]



from ninja import Router, Query
from django.http import HttpResponse
from hydrothings.engine import SensorThingsRequest
from hydrothings.schemas import Filters
from hydrothings.utils import entity_or_404, list_response_codes, get_response_codes
from .schemas import FeatureOfInterestPostBody, FeatureOfInterestPatchBody, FeatureOfInterestListResponse, \
    FeatureOfInterestGetResponse


router = Router(tags=['Features Of Interest'])


@router.get(
    '/FeaturesOfInterest',
    response=list_response_codes(FeatureOfInterestListResponse),
    url_name='list_feature_of_interest'
)
def get_features_of_interest(request: SensorThingsRequest, filters: Filters = Query(...)):
    """
    Get a collection of Feature of Interest entities.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/feature-of-interest/properties" target="_blank">\
      Feature of Interest Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/feature-of-interest/relations" target="_blank">\
      Feature of Interest Relations</a>
    """

    response = request.engine.list(**filters.dict())

    return 200, response


@router.get(
    '/FeaturesOfInterest({feature_of_interest_id})',
    response=get_response_codes(FeatureOfInterestGetResponse),
    by_alias=True
)
def get_feature_of_interest(request, feature_of_interest_id: str):
    """
    Get a Feature of Interest entity.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/feature-of-interest/properties" target="_blank">\
      Feature of Interest Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/feature-of-interest/relations" target="_blank">\
      Feature of Interest Relations</a>
    """

    response = request.engine.get(entity_id=feature_of_interest_id)

    return entity_or_404(response, feature_of_interest_id)


@router.post(
    '/FeaturesOfInterest',
    response={201: None}
)
def create_feature_of_interest(
        request: SensorThingsRequest,
        response: HttpResponse,
        feature_of_interest: FeatureOfInterestPostBody
):
    """
    Create a new Feature of Interest entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/feature-of-interest/properties" target="_blank">\
      Feature of Interest Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/feature-of-interest/relations" target="_blank">\
      Feature of Interest Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/create-entity" target="_blank">\
      Create Entity</a>
    """

    feature_of_interest_id = request.engine.create(
        entity_body=feature_of_interest
    )

    response['location'] = request.engine.get_ref(
        entity_id=feature_of_interest_id
    )

    return 201, None


@router.patch(
    '/FeaturesOfInterest({feature_of_interest_id})',
    response={204: None}
)
def update_feature_of_interest(
        request: SensorThingsRequest,
        feature_of_interest_id: str,
        feature_of_interest: FeatureOfInterestPatchBody
):
    """
    Update an existing Feature of Interest entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/feature-of-interest/properties" target="_blank">\
      Feature of Interest Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/feature-of-interest/relations" target="_blank">\
      Feature of Interest Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/update-entity" target="_blank">\
      Update Entity</a>
    """

    request.engine.update(
        entity_id=feature_of_interest_id,
        entity_body=feature_of_interest
    )

    return 204, None


@router.delete(
    '/FeaturesOfInterest({feature_of_interest_id})',
    response={204: None}
)
def delete_feature_of_interest(request: SensorThingsRequest, feature_of_interest_id: str):
    """
    Delete a Feature of Interest entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/delete-entity" target="_blank">\
      Delete Entity</a>
    """

    request.engine.delete(
        entity_id=feature_of_interest_id
    )

    return 204, None



import inspect
import hydrothings.components as core_components
import hydrothings.schemas as core_schemas
from pydantic import BaseModel


def nested_entities_check(value, field):
    """
    Validation for nested components.

    Runs validation on nested components in request bodies to avoid circular relationships in the API documentation.

    :param value: The input value.
    :param field: The field being validated.
    :return value: The output value.
    """

    nested_class_name = field.field_info.extra.get('nested_class')

    if nested_class_name:
        nested_class = getattr(core_components, nested_class_name)

        if isinstance(value, list):
            value = [
                nested_class(**sub_value.dict()) if isinstance(sub_value, core_schemas.NestedEntity) else sub_value
                for sub_value in value
            ]

        elif isinstance(value, core_schemas.NestedEntity):
            value = nested_class(**value.dict())

    return value


def whitespace_to_none(value):
    """
    Validation for whitespace values.

    Checks if string values are blank or only whitespace and converts them to NoneType.

    :param value: The input value.
    :return value: The output value.
    """

    if isinstance(value, str) and (value == '' or value.isspace()):
        return None

    return value


def allow_partial(*fields):
    """
    Declares fields of a Pydantic model as optional.

    This decorator can be used to update required fields of a Pydantic model to be optional. This can be used in
    conjunction with exclude_unset to allow for missing fields, but still adhere to not_null requirements, such as when
    partially updating a datastore.

    :param fields: All fields which should be made optional. If none are passed, all fields will be made optional.
    :return dec: The optional class decorator.
    """

    def dec(_cls):
        for field in fields:
            _cls.__fields__[field].required = False
        return _cls

    if fields and inspect.isclass(fields[0]) and issubclass(fields[0], BaseModel):
        cls = fields[0]
        fields = cls.__fields__
        return dec(cls)

    return dec



from ninja import Router, Query
from django.http import HttpResponse
from hydrothings.engine import SensorThingsRequest
from hydrothings.schemas import Filters
from hydrothings.utils import entity_or_404, list_response_codes, get_response_codes
from .schemas import ObservationPostBody, ObservationPatchBody, ObservationListResponse, ObservationGetResponse


router = Router(tags=['Observations'])


@router.get(
    '/Observations',
    response=list_response_codes(ObservationListResponse),
    by_alias=True,
    url_name='list_observation',
    exclude_none=True
)
def get_observations(request: SensorThingsRequest, filters: Filters = Query(...)):
    """
    Get a collection of Observation entities.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/properties" target="_blank">\
      Observation Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/relations" target="_blank">\
      Observation Relations</a>
    """

    response = request.engine.list(**filters.dict())

    return 200, response


@router.get(
    '/Observations({observation_id})',
    response=get_response_codes(ObservationGetResponse),
    by_alias=True
)
def get_observation(request: SensorThingsRequest, observation_id: str):
    """
    Get an Observation entity.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/properties" target="_blank">\
      Observation Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/relations" target="_blank">\
      Observation Relations</a>
    """

    response = request.engine.get(entity_id=observation_id)

    return entity_or_404(response, observation_id)


@router.post(
    '/Observations',
    response={201: None}
)
def create_observation(request: SensorThingsRequest, response: HttpResponse, observation: ObservationPostBody):
    """
    Create a new Observation entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/properties" target="_blank">\
      Observation Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/relations" target="_blank">\
      Observation Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/create-entity" target="_blank">\
      Create Entity</a>
    """

    observation_id = request.engine.create(
        entity_body=observation
    )

    response['location'] = request.engine.get_ref(
        entity_id=observation_id
    )

    return 201, None


@router.patch(
    '/Observations({observation_id})',
    response={204: None}
)
def update_observation(request: SensorThingsRequest, observation_id: str, observation: ObservationPatchBody):
    """
    Update an existing Observation entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/properties" target="_blank">\
      Observation Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observation/relations" target="_blank">\
      Observation Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/update-entity" target="_blank">\
      Update Entity</a>
    """

    request.engine.update(
        entity_id=observation_id,
        entity_body=observation
    )

    return 204, None


@router.delete(
    '/Observations({observation_id})',
    response={204: None}
)
def delete_observation(request: SensorThingsRequest, observation_id: str):
    """
    Delete a Observation entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/delete-entity" target="_blank">\
      Delete Entity</a>
    """

    request.engine.delete(
        entity_id=observation_id
    )

    return 204, None



from ninja import Router, Query
from django.http import HttpResponse
from hydrothings.engine import SensorThingsRequest
from hydrothings.schemas import Filters
from hydrothings.utils import entity_or_404, list_response_codes, get_response_codes
from .schemas import ObservedPropertyPostBody, ObservedPropertyPatchBody, ObservedPropertyListResponse, \
    ObservedPropertyGetResponse


router = Router(tags=['Observed Properties'])


@router.get(
    '/ObservedProperties',
    response=list_response_codes(ObservedPropertyListResponse),
    by_alias=True,
    url_name='list_observed_property',
    exclude_none=True
)
def get_observed_properties(request: SensorThingsRequest, filters: Filters = Query(...)):
    """
    Get a collection of Observed Property entities.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observed-property/properties" target="_blank">\
      Observed Property Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observed-property/relations" target="_blank">\
      Observed Property Relations</a>
    """

    response = request.engine.list(**filters.dict())

    return 200, response


@router.get(
    '/ObservedProperties({observed_property_id})',
    response=get_response_codes(ObservedPropertyGetResponse),
    by_alias=True
)
def get_observed_property(request, observed_property_id: str):
    """
    Get an Observed Property entity.

    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observed-property/properties" target="_blank">\
      Observed Property Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observed-property/relations" target="_blank">\
      Observed Property Relations</a>
    """

    response = request.engine.get(entity_id=observed_property_id)

    return entity_or_404(response, observed_property_id)


@router.post(
    '/ObservedProperties',
    response={201: None}
)
def create_observed_property(
        request: SensorThingsRequest,
        response: HttpResponse,
        observed_property: ObservedPropertyPostBody
):
    """
    Create a new Observed Property entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observed-property/properties" target="_blank">\
      Observed Property Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observed-property/relations" target="_blank">\
      Observed Property Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/create-entity" target="_blank">\
      Create Entity</a>
    """

    observed_property_id = request.engine.create(
        entity_body=observed_property
    )

    response['location'] = request.engine.get_ref(
        entity_id=observed_property_id
    )

    return 201, None


@router.patch(
    '/ObservedProperties({observed_property_id})',
    response={204: None}
)
def update_observed_property(
        request: SensorThingsRequest,
        observed_property_id: str,
        observed_property: ObservedPropertyPatchBody
):
    """
    Update an existing Observed Property entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observed-property/properties" target="_blank">\
      Thing Properties</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/datamodel/observed-property/relations" target="_blank">\
      Thing Relations</a> -
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/update-entity" target="_blank">\
      Update Entity</a>
    """

    request.engine.update(
        entity_id=observed_property_id,
        entity_body=observed_property
    )

    return 204, None


@router.delete(
    '/ObservedProperties({observed_property_id})',
    response={204: None}
)
def delete_observed_property(request: SensorThingsRequest, observed_property_id: str):
    """
    Delete an Observed Property entity.

    Links:
    <a href="http://www.opengis.net/spec/iot_sensing/1.1/req/create-update-delete/delete-entity" target="_blank">\
      Delete Entity</a>
    """

    request.engine.delete(
        entity_id=observed_property_id
    )

    return 204, None



from ninja import Router
from django.urls import reverse
from .schemas import ServerRootResponse
from hydrothings import settings


router = Router()


@router.get(
    '',
    include_in_schema=False,
    by_alias=True,
    response=ServerRootResponse
)
def get_root(request):
    """
    Get SensorThings server capabilities.
    """

    host_url = request.get_host()
    response = {
        'server_settings': {
            'conformance': settings.ST_CONFORMANCE
        },
        'server_capabilities': [
            {
                'name': capability['NAME'],
                'url': host_url + reverse(f"api-{settings.ST_VERSION}:{capability['VIEW']}")
            } for capability in settings.ST_CAPABILITIES
        ]
    }

    return response
